# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from enum import IntEnum

from .dispatch_rule import DispatchRule


class HWPCDepthLevel(IntEnum):
    """
    Enumeration that specify which report level use to group by the reports
    """

    TARGET = -1
    ROOT = 0
    SOCKET = 1
    CORE = 2


class HWPCDispatchRule(DispatchRule):
    """
    Group by rule for HWPC report
    """
    def __init__(self, depth, primary=False):
        """
        :param depth:
        :type depth: HWPCDepthLevel
        """
        DispatchRule.__init__(self, primary)
        self.depth = depth
        self.fields = self._set_field()

    def _set_field(self):
        if self.depth == HWPCDepthLevel.TARGET:
            return ['target']

        return ['sensor', 'socket', 'core'][:(self.depth + 1)]

    def get_formula_id(self, report):
        """
        See :meth:`DispatchRule.extract <powerapi.dispatch_rule.abstract_dispatch_rule.DispatchRule.extract>`
        """
        if self.depth == HWPCDepthLevel.TARGET:
            return [(report.target,)]

        if self.depth == HWPCDepthLevel.ROOT:
            return [(report.sensor,)]

        non_shared_group = _extract_non_shared_group(report)

        if self.depth == HWPCDepthLevel.SOCKET:
            id_list = []
            for socket_report in non_shared_group.items():
                id_list.append((report.sensor, socket_report[0]))
            return id_list

        if self.depth == HWPCDepthLevel.CORE:
            id_list = []
            for socket_report in non_shared_group.items():
                for core_report in socket_report[1].items():
                    id_list.append((report.sensor, socket_report[0],
                                    core_report[0]))
            return id_list

        return []


def _number_of_core_per_socket(group):
    """
    Compute the number of core per socket in this group
    :param group: group must be a valide group (composed by socket, core, event)
    :type group: Dict
    :rtype: int : the number of core per socket in this group
    """
    return len(list(group.values())[0])


def _extract_non_shared_group(report):
    """
    extract a non shared group form the given report.
    A shared group is a group that contains events that are shared between
    multiple cores (like RAPL or PCU)

    :rtype:{str:HWPCReportSocket}: a group containing ReportSocket
    """
    biggest_group = None
    maximum_number_of_core = -1
    for _, group in report.groups.items():
        number_of_core = _number_of_core_per_socket(group)
        if number_of_core > maximum_number_of_core:
            maximum_number_of_core = number_of_core
            biggest_group = group
    return biggest_group



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging

import pytest

from thespian.actors import Actor, ActorExitRequest

from powerapi.message import StartMessage, OKMessage, ErrorMessage
from powerapi.actor import Actor as PowerapiActor, InitializationException
from powerapi.formula import DomainValues

LOGGER_NAME = 'thespian_test_logger'


@pytest.fixture
def logger(system, dummy_pipe_in):
    """
    fixture that return a DummyActor

    A DummyActor is an actor that send every received message to the pytest process through a pipe
    This type of actor is usefull to unit test actor that interact with other actors.
    It may tests if the tested actor send the correct message to the actor it must interact with
    """
    logger_actor = system.createActor(DummyActor, globalName=LOGGER_NAME)
    system.tell(logger_actor, DummyStartMessage('system', 'logger', dummy_pipe_in))
    yield logger_actor
    system.tell(logger_actor, ActorExitRequest())


class DummyStartMessage(StartMessage):
    """
    Message used to start a DummyActor
    :param pipe: pipe used to send message to pytest process
    """
    def __init__(self, sender_name, name, pipe):
        StartMessage.__init__(self, sender_name, name)
        self.pipe = pipe


class DummyActor(Actor):
    """
    Actor that forward all start message (except StartMessage) into a pipe to the test process
    """
    def __init__(self):
        Actor.__init__(self)
        self.pipe = None
        self.name = None

    def receiveMessage(self, message, sender):
        """
        when receive a message, send it to the pytest process through a pipe
        """
        if isinstance(message, DummyStartMessage):
            self.pipe = message.pipe
            self.name = message.name
        else:
            self.pipe.send((self.name, message))
            logging.debug('receive : ' + str(message), extra={'actor_name': self.name})


class DummyFormulaActor(Actor):
    """
    A fake FormulaActor that is connected to a DummyActor (a fake pusher)
    """
    def __init__(self):
        Actor.__init__(self)
        self.name = None
        self.fake_puller = None

    def receiveMessage(self, message, sender):
        """
        When receiving a message :
        if its a Start message containing initialization values : initailize the DummyFormula
        if its an ActorExitRequest : notify the pusher that the Dummyformula die
        if its an other type of message : forward it to the fake pusher
        """
        logging.debug('receive : ' + str(message), extra={'actor_name': self.name})
        if isinstance(message, StartMessage):
            self.name = message.name
            self.fake_puller = self.createActor(DummyActor, globalName=message.values.pushers['fake_pusher'])
            self.send(self.fake_puller, message)
            self.send(sender, OKMessage(self.name))
        elif isinstance(message, ActorExitRequest):
            self.send(self.fake_puller, 'dead')
        else:
            self.send(self.fake_puller, message)

    @staticmethod
    def gen_domain_values(device_id, formula_id):
        return DomainValues(device_id, formula_id)


class CrashException(Exception):
    """
    Exception raised by formla to make it crash
    """


class CrashFormulaActor(DummyFormulaActor):
    """
    Formula that crash after receiving 3 messages
    """
    def __init__(self):
        DummyFormulaActor.__init__(self)
        self.report_received = 0

    def receiveMessage(self, message, sender):
        logging.debug('receive : ' + str(message), extra={'actor_name': self.name})
        if isinstance(message, StartMessage):
            self.name = message.name
            self.fake_puller = self.createActor(DummyActor, globalName=message.values.pushers['fake_pusher'])
            self.send(self.fake_puller, message)
            self.send(sender, OKMessage(self.name))
        elif isinstance(message, ActorExitRequest):
            self.send(self.fake_puller, 'dead')
        else:
            self.report_received += 1
            if self.report_received >= 3:
                self.send(self.fake_puller, 'crash')
                raise CrashException
            self.send(self.fake_puller, message)


class CrashInitFormulaActor(DummyFormulaActor):
    """
    Formula answer ErrorMessage to StartMessage
    Like DummyFormulaActor, it will forward received Message to fake pusher
    """
    def __init__(self):
        DummyFormulaActor.__init__(self)
        self.report_received = 0

    def receiveMessage(self, message, sender):
        if isinstance(message, StartMessage):
            self.name = message.name
            self.fake_puller = self.createActor(DummyActor, globalName=message.values.pushers['fake_pusher'])
            self.send(sender, ErrorMessage(self.name, 'formula crashed'))
        elif isinstance(message, ActorExitRequest):
            self.send(self.fake_puller, 'dead')
        else:
            self.send(self.fake_puller, message)


class DummyPowerapiActor(PowerapiActor):
    """
    Actor that have the same API than a basic powerapi actor
    """

    def __init__(self):
        PowerapiActor.__init__(self, StartMessage)


class CrashInitActor(DummyPowerapiActor):
    """
    Basic powerapi actor that crash at initialisation
    """
    def _initialization(self, msg):
        raise InitializationException('error')


class CrashActor(DummyPowerapiActor):
    """
    Basic powerapi actor that crash 2s after initialization
    """
    def _initialization(self, _):
        self.wakeupAfter(2)

    def receiveMsg_WakeupMessage(self, message, sender):
        """
        crash after being waked up by system, 2s after initialization
        """
        raise CrashException()



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from thespian.actors import ActorAddress, ActorExitRequest

from powerapi.actor import Actor, InitializationException
from powerapi.message import PusherStartMessage, EndMessage
from powerapi.database import DBError
from powerapi.report import PowerReport, BadInputData
from powerapi.exception import PowerAPIExceptionWithMessage, PowerAPIException


class PusherActor(Actor):
    """
    PusherActor class

    The Pusher allow to save Report sent by Formula.
    """

    def __init__(self):
        Actor.__init__(self, PusherStartMessage)
        self.database = None

    def _initialization(self, start_message: PusherStartMessage):
        Actor._initialization(self, start_message)
        self.database = start_message.database

        try:
            self.database.connect()
        except DBError as error:
            raise InitializationException(error.msg) from error

    def receiveMsg_PowerReport(self, message: PowerReport, _: ActorAddress):
        """
        When receiving a PowerReport save it to database
        """
        self.log_debug('received message ' + str(message))
        try:
            self.database.save(message)
            self.log_debug(str(message) + 'saved to database')
        except BadInputData as exn:
            log_line = 'BadinputData exception raised for report' + str(exn.input_data)
            log_line += ' with message : ' + exn.msg
            self.log_warning(log_line)
        except PowerAPIExceptionWithMessage as exn:
            log_line = 'exception ' + str(exn) + 'was raised while trying to save ' + str(message)
            log_line += 'with message : ' + str(exn.msg)
            self.log_warning(log_line)
        except PowerAPIException as exn:
            self.log_warning('exception ' + str(exn) + 'was raised while trying to save ' + str(message))

    def receiveMsg_EndMessage(self, message: EndMessage, _: ActorAddress):
        """
        When receiving an EndMessage notify the ActorSystem and Kill itself
        """
        self.log_debug('received message ' + str(message))
        self.send(self.parent, EndMessage(self.name))
        self.send(self.myAddress, ActorExitRequest())



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Tuple
from enum import IntEnum

from powerapi.report import PowerReport

from .dispatch_rule import DispatchRule


class PowerDepthLevel(IntEnum):
    """
    Enumeration that specify which report level use to group by the reports
    """

    TARGET = -1
    SENSOR = 0
    SOCKET = 1
    CORE = 2


def extract_id_from_report(report: PowerReport, depth: PowerDepthLevel) -> Tuple:
    """
    :return: a report id generated from the report and the given depth
    """
    if depth == PowerDepthLevel.TARGET:
        return (report.target,)

    if depth == PowerDepthLevel.SENSOR:
        return (report.sensor,)

    if depth == PowerDepthLevel.SOCKET:
        return extract_id_from_report(report, depth - 1) + (report.metadata['socket'],)

    return extract_id_from_report(report, depth - 1) + (report.metadata['core'],)


class PowerDispatchRule(DispatchRule):
    """
    Group by rule for HWPC report
    """
    def __init__(self, depth: PowerDepthLevel, primary=False):
        """
        :param depth:
        :type depth: HWPCDepthLevel
        """
        DispatchRule.__init__(self, primary)
        self.depth = depth
        self.fields = self._set_field()

    def _set_field(self):
        if self.depth == PowerDepthLevel.TARGET:
            return ['target']

        return ['sensor', 'socket', 'core'][:(self.depth + 1)]

    def get_formula_id(self, report):
        return [extract_id_from_report(report, self.depth)]



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import re
import logging

try:
    from libvirt import openReadOnly, libvirtError
except ImportError:
    logging.getLogger().info("libvirt-python is not installed.")

    class LibvirtException(Exception):
        """"""
        def __init__(self, _):
            Exception.__init__(self)

    libvirtError = LibvirtException
    openReadOnly = None

from powerapi.report_modifier.report_modifier import ReportModifier


class LibvirtMapper(ReportModifier):
    """
    Report modifier which modifi target with libvirt id by open stak uuid
    """

    def __init__(self, uri: str, regexp: str):
        self.regexp = re.compile(regexp)
        daemon_uri = None if uri == '' else uri
        self.libvirt = openReadOnly(daemon_uri)

    def modify_report(self, report):
        result = re.match(self.regexp, report.target)
        if result is not None:
            domain_name = result.groups(0)[0]
            try:
                domain = self.libvirt.lookupByName(domain_name)
                report.metadata["domain_id"] = domain.UUIDString()
            except libvirtError:
                pass
        return report



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import pytest

from influxdb import InfluxDBClient


INFLUX_URI = 'localhost'
INFLUX_PORT = 8086
INFLUX_DBNAME = 'acceptation_test'


@pytest.fixture()
def influx_database(influxdb_content):
    """
    connect to a local influx database (localhost:8086) and store data contained in the list influxdb_content
    after test end, delete the data
    """
    client = InfluxDBClient(host=INFLUX_URI, port=INFLUX_PORT)
    _delete_db(client, INFLUX_DBNAME)
    _init_db(client, influxdb_content)
    yield client
    _delete_db(client, INFLUX_DBNAME)


def _init_db(client, content):

    if content != []:
        client.create_database(INFLUX_DBNAME)
        client.switch_database(INFLUX_DBNAME)
        client.write_points(content)


def _delete_db(client, db_name):
    client.drop_database(db_name)
    client.close()


def get_all_reports(client, db_name):
    """
    get all points stored in the database during test execution
    """
    client.switch_database(db_name)
    result = client.query('SELECT * FROM "power_consumption"')
    return list(result.get_points())



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from powerapi.database import BaseDB, DBError
from powerapi.report import Report


class FakeDBError(Exception):
    """
    Exception raised to crash the db
    """


class FakeDB(BaseDB):
    """
    A fake database that send information throug a pipe when its api is used
    """

    def __init__(self, content=[], pipe=None, *args, **kwargs):
        BaseDB.__init__(self, Report)
        self._content = content
        self.pipe = pipe

    def connect(self):
        """
        send the string connected through the pipe
        """
        if self.pipe is not None:
            self.pipe.send('connected')

    def iter(self, stream_mode):
        return self._content.__iter__()

    def save(self, report):
        """
        send the saved report through the pipe
        """
        if self.pipe is not None:
            self.pipe.send(report)

    def save_many(self, reports):
        """
        send the saved reports through the pipe
        """
        if self.pipe is not None:
            self.pipe.send(reports)


class SilentFakeDB(BaseDB):
    """
    An empty Database that don't send information through the pipe
    """
    def __init__(self, content=[], pipe=None, *args, **kwargs):
        BaseDB.__init__(self, Report)
        self._content = content

    def connect(self):
        pass

    def iter(self, stream_mode):
        return self._content.__iter__()

    def save(self, report):
        pass

    def save_many(self, reports):
        pass


class CrashDB(BaseDB):
    """
    FakeDB that crash when using its connect method
    """
    def __init__(self, *args, **kwargs):
        BaseDB.__init__(self, Report)

    def connect(self):
        raise DBError('crash')


def define_database(database):
    """
    Decorator to set the _database
    attribute for individual tests.

    ! If you use this decorator, you need to insert handler in  pytest_generate_tests function !
    ! see tests/unit/test_puller.py::pytest_generate_tests for example  !
    """
    def wrap(func):
        setattr(func, '_database', database)
        return func
    return wrap


def define_report_type(report_type):
    """
    Decorator to set the _report_type
    attribute for individuel tests.

    ! If you use this decorator, you need to insert handler in  pytest_generate_tests function !
    ! see tests/unit/test_puller.py::pytest_generate_tests for example  !
    """
    def wrap(func):
        setattr(func, '_report_type', report_type)
        return func
    return wrap



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Type, Tuple

from powerapi.formula.formula_actor import FormulaActor, DomainValues
from powerapi.message import FormulaStartMessage


class CpuDramDomainValues(DomainValues):
    """
    Socket values for device with socket and core domains
    """

    def __init__(self, device_id: str, formula_id: Tuple):
        DomainValues.__init__(self, device_id, formula_id)

        self.socket = None if len(formula_id) <= 1 else int(formula_id[1])
        self.core = None if len(formula_id) <= 2 else int(formula_id[2])


class AbstractCpuDramFormula(FormulaActor):
    """
    Formula that handle CPU or DRAM related data
    It can be launched to handle data from a specific part of a cpu (whole cpu, a socket or a core)
    """
    def __init__(self, start_message_cls: Type[FormulaStartMessage]):
        FormulaActor.__init__(self, start_message_cls)
        self.socket = None
        self.core = None

    def _initialization(self, start_message: FormulaStartMessage):
        FormulaActor._initialization(self, start_message)
        self.socket = start_message.domain_values.socket
        self.core = start_message.domain_values.core

    @staticmethod
    def gen_domain_values(device_id: str, formula_id: Tuple) -> CpuDramDomainValues:
        return CpuDramDomainValues(device_id, formula_id)



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging
try:
    from opentsdb import TSDBClient
except ImportError:
    logging.getLogger().info("opentsdb-py is not installed.")

from typing import List, Type

from powerapi.report import PowerReport, Report
from .base_db import BaseDB, DBError


class CantConnectToOpenTSDBException(DBError):
    """
    Exception raised to notify that connection to the opentsdb database is impossible
    """


class OpenTSDB(BaseDB):
    """
    OpenTSDB class herited from BaseDB

    Allow to handle an OpenTSDB database to save PowerReport.
    """

    def __init__(self, report_type: Type[Report], host: str, port, metric_name: str):
        """
        :param host:             host of the OpenTSDB server
        :param port:            port of the OpenTSDB server

        :param metric_name:         mectric name to store


        :param report_type:        type of report handled by this database

        """
        BaseDB.__init__(self, report_type)
        self.host = host
        self.port = port
        self.metric_name = metric_name

        self.client = None

    def __iter__(self):
        raise NotImplementedError()

    def connect(self):
        """
        Override from BaseDB.

        Create the connection to the openTSDB database with the current
        configuration (hostname/port), then check if the connection has
        been created without failure.

        """
        # close connection if reload
        if self.client is not None:
            self.client.close()
            self.client.wait()

        self.client = TSDBClient(host=self.host, port=self.port)

        if not self.client.is_connected() and not self.client.is_alive():
            raise CantConnectToOpenTSDBException('connexion error')

    def save(self, report: PowerReport):
        """
        Override from BaseDB

        :param report: Report to save
        """
        self.client.send(self.metric_name, report.power, timestamp=int(report.timestamp.timestamp()),
                         host=report.target)

    def save_many(self, reports: List[Report]):
        """
        Save a batch of data

        :param reports: Batch of data.
        """

        for report in reports:
            self.save(report)



import pytest


@pytest.fixture
def procfs_timeline():
    """
    Timeline of procfs report for the tests
    """

    return [
        {
            "timestamp": "2021-09-14T12:37:37.168817",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:37.669237",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:38.170142",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:38.670338",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:39.171321",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:39.671572",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:40.172503",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:40.672693",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:41.173552",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:41.673815",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:42.174560",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:42.674690",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:43.175441",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:43.675743",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:44.176551",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:44.677307",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:45.178049",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:45.678310",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:46.179120",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:46.679308",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:47.180223",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:47.680468",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:48.181316",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:48.681683",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:49.182522",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:49.682731",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:50.183680",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:50.683812",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:51.184792",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:51.685027",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:52.185709",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:52.686065",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.610000000000014,
        },
        {
            "timestamp": "2021-09-14T12:37:53.186929",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.36,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:53.687190",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.35,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:54.188031",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.35,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.600000000000012,
        },
        {
            "timestamp": "2021-09-14T12:37:54.688674",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.35,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.59000000000001,
        },
        {
            "timestamp": "2021-09-14T12:37:55.189489",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.35,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.59000000000001,
        },
        {
            "timestamp": "2021-09-14T12:37:55.690299",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.35,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.59000000000001,
        },
        {
            "timestamp": "2021-09-14T12:37:56.191124",
            "sensor": "formula_group",
            "target": ["firefox_cgroup", "emacs_cgroup",
                       "zsh_cgroup", "mongo_cgroup"],
            "usage": {
                "firefox_cgroup": 8.35,
                "emacs_cgroup": 5.52,
                "zsh_cgroup": 0.01,
                "mongo_cgroup": 0.64,
            },
            "global_cpu_usage": 27.59000000000001,
        },
    ]


@pytest.fixture
def power_timeline():
    """
    Timeline of procfs report for the tests
    """

    return [
        {
            "timestamp": "2021-09-14T12:37:37.168817",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:37.669237",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:38.170142",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:38.670338",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:39.171321",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:39.671572",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:40.172503",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:40.672693",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:41.173552",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:41.673815",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:42.174560",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:42.674690",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:43.175441",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:43.675743",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:44.176551",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:44.677307",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:45.178049",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:45.678310",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:46.179120",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:46.679308",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:47.180223",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:47.680468",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:48.181316",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:48.681683",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:49.182522",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:49.682731",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:50.183680",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:50.683812",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:51.184792",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:51.685027",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:52.185709",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:52.686065",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:53.186929",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:53.687190",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:54.188031",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:54.688674",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:55.189489",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:55.690299",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },
        {
            "timestamp": "2021-09-14T12:37:56.191124",
            "sensor": "formula_group",
            "target": "all",
            "power": 42,
        },

    ]



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import asyncio
from typing import Type, List
import json


from powerapi.utils import JsonStream
from powerapi.report import Report
from .base_db import IterDB, BaseDB, DBError

BUFFER_SIZE = 4096
SOCKET_TIMEOUT = 0.5


class SocketDB(BaseDB):
    """
    Database that act as a server that expose a socket where data source will push data
    """

    def __init__(self, report_type: Type[Report], port: int):
        BaseDB.__init__(self, report_type)
        self.asynchrone = True
        self.queue = None
        self.port = port
        self.server = None

    async def connect(self):
        self.queue = asyncio.Queue()
        self.server = await asyncio.start_server(self._gen_server_callback(), host='0.0.0.0', port=self.port)

    async def stop(self):
        """
        stop server connection
        """
        self.server.close()
        await self.server.wait_closed()

    def iter(self, stream_mode):
        return IterSocketDB(self.report_type, stream_mode, self.queue)

    def _gen_server_callback(self):
        async def callback(stream_reader, _):
            stream = JsonStream(stream_reader)
            count = 0  # If 10 times in a row we don't have a full message we stop
            while True:
                json_str = await stream.read_json_object()
                if json_str is None:
                    if count > 10:
                        break
                    count += 1
                    continue
                count = 0
                await self.queue.put(json_str)
                # self.queue.put(json_str)

        return callback

    def __iter__(self):
        raise DBError('Socket db don\'t support __iter__ method')

    def save(self, report: Report):
        raise DBError('Socket db don\'t support save method')

    def save_many(self, reports: List[Report]):
        raise DBError('Socket db don\'t support save_many method')


class IterSocketDB(IterDB):
    """
    iterator connected to a socket that receive report from a sensor
    """

    def __init__(self, report_type, stream_mode, queue):
        """
        """
        IterDB.__init__(self, None, report_type, stream_mode)

        self.queue = queue

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            json_str = await asyncio.wait_for(self.queue.get(), 2)
            # json = self.queue.get_nowait()
            # self.queue.get()
            report = self.report_type.from_json(json.loads(json_str))
            return report
        # except Empty:
        except asyncio.TimeoutError:
            return None



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import asyncio

from thespian.actors import ActorExitRequest

from powerapi.actor import TimedActor, InitializationException
from powerapi.report import BadInputData
from powerapi.database import DBError
from powerapi.message import PullerStartMessage, EndMessage


class PullerActor(TimedActor):
    """
    Actor used to pull data from sources.

    A puller Actor is configured to pull data from one type of sources
    """
    def __init__(self):
        TimedActor.__init__(self, PullerStartMessage, 0.05)

        self.database = None
        self.report_filter = None
        self.stream_mode = None
        self.database_it = None
        self.report_modifier_list = None
        self.loop = None

        self._number_of_message_before_sleeping = 10

    def _initialization(self, start_message: PullerStartMessage):
        TimedActor._initialization(self, start_message)

        self.database = start_message.database
        self.report_filter = start_message.report_filter
        self.stream_mode = start_message.stream_mode
        self.report_modifier_list = start_message.report_modifier_list

        self._database_connection()
        if not self.report_filter.filters:
            raise InitializationException('filter without rules')

    def _database_connection(self):
        try:
            if not self.database.asynchrone:
                self.database.connect()
            else:
                self.loop = asyncio.new_event_loop()
                asyncio.set_event_loop(self.loop)
                self.loop.set_debug(enabled=True)
                logging.basicConfig(level=logging.DEBUG)
                self.loop.run_until_complete(self.database.connect())
            self.database_it = self.database.iter(self.stream_mode)
        except DBError as error:
            raise InitializationException(error.msg) from error

    def _modify_report(self, report):
        for report_modifier in self.report_modifier_list:
            report = report_modifier.modify_report(report)
        return report

    def _launch_task(self):
        """
        Initialize the database and connect all dispatcher to the
        socket_interface
        """
        for _ in range(self._number_of_message_before_sleeping):
            try:
                raw_report = self._pull_database()
                report = self._modify_report(raw_report)
                dispatchers = self.report_filter.route(report)
                for dispatcher in dispatchers:
                    self.log_debug('send report ' + str(report) + 'to ' + str(dispatcher))
                    self.send(dispatcher, report)
                self.wakeupAfter(self._time_interval)
                return
            except StopIteration:
                if self.stream_mode:
                    self.wakeupAfter(self._time_interval)
                    return
                self.log_info('input source empty, stop system')
                self._terminate()
                return
            except BadInputData as exn:
                log_line = 'BadinputData exception raised for input data' + str(exn.input_data)
                log_line += ' with message : ' + exn.msg
                self.log_warning(log_line)

    def _terminate(self):
        self.send(self.parent, EndMessage(self.name))
        for _, dispatcher in self.report_filter.filters:
            self.send(dispatcher, EndMessage(self.name))
        self.send(self.myAddress, ActorExitRequest())

    def _pull_database(self):
        if self.database.asynchrone:
            report = self.loop.run_until_complete(self.database_it.__anext__())
            # print(report)
            if report is not None:
                return report
            else:
                raise StopIteration()
        else:
            return next(self.database_it)



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import collections.abc as collections
import datetime


def timestamp_to_datetime(timestamp):
    """
    Create datetime from a timestamp value

    :param int timestamp:
    :rtype: datetime.datetime
    """
    return datetime.datetime.fromtimestamp(timestamp / 1000)


def datetime_to_timestamp(date):
    """
    Cast a datetime object to a simple timestamp

    :param datetime.datetime date:
    :rtype: int
    """
    return int(datetime.datetime.timestamp(date) * 1000)


def dict_merge(dict1, dict2):
    """
    Recursive dict merge, act like dict.update() but update
    all level and not only top-level.

    :param dict dict1:
    :param dict dict2:

    """
    for key, value in dict2.items():
        if (key in dict1 and isinstance(dict1[key], dict) and isinstance(dict2[key], collections.Mapping)):
            dict_merge(dict1[key], dict2[key])
        else:
            dict1[key] = value



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

from datetime import datetime
from typing import Dict, List, Any

from .report import Report


class ControlReport(Report):
    """
    EventReport stores information about an event.
    This is useful to control external tools via a producer/consumer job queue.
    """

    def __init__(self, timestamp: datetime, sensor: str, target: str, action: str, parameters: List, metadata: Dict[str, Any] = {}):
        """
        Initialize a Control Event report using the given parameters.
        :param timestamp: Report timestamp
        :param sensor: Sensor name
        :param target: Target name
        :param action: Action name
        :param parameters: Parameter values
        """
        Report.__init__(self, timestamp, sensor, target, metadata)
        self.action = action
        self.parameters = parameters

    def __repr__(self) -> str:
        return 'ControlReport(%s, %s, %s, %s, %s, %s)' % (self.timestamp, self.sensor, self.target, self.action, self.parameters, str(self.metadata))

    @staticmethod
    def from_json(data: Dict) -> ControlReport:
        """
        :return: a dictionary, that can be converted into json format, from a given ControlReport
        """
        metadata = {} if 'metadata' not in data else data['metadata']
        return ControlReport(data['timestamp'], data['sensor'], data['target'], data['action'], data['parameters'], metadata)

    @staticmethod
    def from_mongodb(data: Dict) -> ControlReport:
        """
        :return: a ControlReport from a dictionary pulled from mongodb
        """
        return ControlReport.from_json(data)

    @staticmethod
    def to_mongodb(report: ControlReport) -> Dict:
        """
        :return: a dictionary, that can be stored into a mongodb, from a given ControlReport
        """
        return report.__dict__



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Dict, Type, Tuple

from thespian.actors import ActorAddress, ActorExitRequest

from powerapi.actor import Actor
from powerapi.message import FormulaStartMessage, EndMessage


class FormulaValues:
    """
    values used to initialize formula actor
    """
    def __init__(self, pushers: Dict[str, ActorAddress]):
        self.pushers = pushers


class DomainValues:
    """
    values that describe the device that the formula compute the power consumption for
    """
    def __init__(self, device_id: str, formula_id: Tuple):
        self.device_id = device_id
        self.sensor = formula_id[0]


class FormulaActor(Actor):
    """
    Abstract actor class used to implement formula actor that compute power consumption of a device from Reports
    """
    def __init__(self, start_message_cls: Type[FormulaStartMessage]):
        Actor.__init__(self, start_message_cls)
        self.pushers: Dict[str, ActorAddress] = None
        self.device_id = None
        self.sensor = None

    def _initialization(self, start_message: FormulaStartMessage):
        Actor._initialization(self, start_message)
        self.pushers = start_message.values.pushers
        self.device_id = start_message.domain_values.device_id
        self.sensor = start_message.domain_values.sensor

    def receiveMsg_EndMessage(self, message: EndMessage, _: ActorAddress):
        """
        when receiving a EndMessage kill itself
        """
        self.log_debug('received message ' + str(message))
        self.send(self.myAddress, ActorExitRequest())

    @staticmethod
    def gen_domain_values(device_id: str, formula_id: Tuple) -> DomainValues:
        """
        generate domain values of the formula from device and formula id
        """
        return DomainValues(device_id, formula_id)



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from multiprocessing import Pipe

import pytest

from thespian.actors import ActorSystem

from powerapi.message import OKMessage, ErrorMessage, PingMessage, EndMessage

from .db import FakeDB, CrashDB
from .dummy_actor import DummyActor, DummyStartMessage, logger
from .actor import system, is_actor_alive

LOGGER_NAME = 'thespian_test_logger'


def recv_from_pipe(pipe, timeout):
    """
    add timeout to function pipe.recv
    """
    if pipe.poll(timeout):
        return pipe.recv()
    else:
        return None, None


class UnknowMessage:
    """
    Message of type unknown.
    """


class AbstractTestActor:
    """
    Basic test that an actor should pass

    To use it with an actor that you create you must implement the method actor and actor_start_message
    test function added to this class must take in parameter the two fixtures system and (actor of started_actor)
    """
    @classmethod
    def teardown_class(cls):
        """
        After all test was executed, shutdown the actor system
        """
        while ActorSystem().listen(0.1) is not None:
            continue
        ActorSystem().shutdown()

    @pytest.fixture
    def dummy_pipe(self):
        """
        pipe used between dummy actor and pytest process
        """
        return Pipe()

    @pytest.fixture
    def dummy_pipe_out(self, dummy_pipe):
        """
        pipe used from pytest process to receive dummy actor information
        """
        return dummy_pipe[0]

    @pytest.fixture
    def dummy_pipe_in(self, dummy_pipe):
        """
        pipe used from dummy actor to send information to pytest process
        """
        return dummy_pipe[1]

    @pytest.fixture
    def actor(self):
        """
        This fixture must return the actor class of the tested actor
        """
        raise NotImplementedError()

    @pytest.fixture
    def actor_start_message(self):
        """
        This fixture must return an instance of start message used to start the tested actor
        """
        raise NotImplementedError()

    @pytest.fixture
    def started_actor(self, system, actor, actor_start_message):
        """
        fixture that return and actor that was started with a StartMessage
        """
        system.ask(actor, actor_start_message)
        return actor

    def test_create_an_actor_and_send_it_PingMessage_must_make_it_answer_OKMessage(self, system, actor):
        msg = system.ask(actor, PingMessage('system'), 0.3)
        print(msg)
        assert isinstance(msg, OKMessage)

    def test_create_an_actor_and_send_it_UnknowMessage_must_make_it_answer_ErrorMessage(self, system, actor):
        msg = system.ask(actor, UnknowMessage(), 0.3)
        print(msg)
        assert isinstance(msg, ErrorMessage)

    def test_send_StartMessage_answer_OkMessage(self, system, actor, actor_start_message):
        msg = system.ask(actor, actor_start_message)
        print(msg)
        assert isinstance(msg, OKMessage)

    def test_send_StartMessage_to_already_started_actor_answer_ErrorMessage(self, system, started_actor, actor_start_message):
        msg = system.ask(started_actor, actor_start_message)
        assert isinstance(msg, ErrorMessage)
        assert msg.error_message == 'Actor already initialized'

    def test_send_EndMessage_to_started_actor_make_it_terminate(self, system, started_actor):
        system.tell(started_actor, EndMessage)
        assert not is_actor_alive(system, started_actor)


def define_database_content(content):
    """
    Decorator used to define database content when using an actor with database

    ! If you use this decorator, you need to insert handler in  pytest_generate_tests function !
    ! see tests/unit/test_puller.py::pytest_generate_tests for example  !
    """
    def wrap(func):
        setattr(func, '_content', content)
        return func
    return wrap


class AbstractTestActorWithDB(AbstractTestActor):
    """
    Base test for actor using a database

    This class add fixtures to communicate with a fake database
    """

    @pytest.fixture
    def pipe(self):
        """
        pipe used to communicate between pytest process and actor
        """
        return Pipe()

    @pytest.fixture
    def pipe_out(self, pipe):
        """
        pipe used from pytest process to receive actor information
        """
        return pipe[0]

    @pytest.fixture
    def pipe_in(self, pipe):
        """
        pipe used from actor process to send information to pytest process
        """
        return pipe[1]

    @pytest.fixture
    def fake_db(self, content, pipe_in):
        """
        Return a FakeDB that will send information through the pipe when using its API
        """
        return FakeDB(content, pipe_in)

    @pytest.fixture
    def crash_db(self):
        """
        Return a FakeDB that will crash when using its connect method
        """
        return CrashDB()

    @pytest.fixture
    def started_actor(self, actor, pipe_out, fake_db, actor_start_message):
        ActorSystem().ask(actor, actor_start_message)
        print(pipe_out.recv())  # remove 'connected' string from Queue
        return actor

    def test_starting_actor_make_it_connect_to_database(self, system, actor, actor_start_message, pipe_out):
        ActorSystem().ask(actor, actor_start_message)
        assert pipe_out.recv() == 'connected'



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging
from typing import List, Type

try:
    from prometheus_client import Gauge
except ImportError:
    logging.getLogger().info("prometheus-client is not installed.")

from powerapi.report import Report
from .prometheus_db import BasePrometheusDB


class DirectPrometheusDB(BasePrometheusDB):
    """
    Database that expose received data as metric in order to be scrapped by a prometheus instance
    Could only be used with a pusher actor
    """
    def __init__(self, report_type: Type[Report], port: int, address: str, metric_name: str, metric_description: str, tags: List[str]):
        """
        :param address:             address that expose the metric
        :param port:
        :param metric_name:
        :param metric_description:  short sentence that describe the metric
        :param tags: metadata used to tag metric
        """
        BasePrometheusDB.__init__(self, report_type, port, address, metric_name, metric_description, tags)

        self.energy_metric = None

        self.current_ts = 0
        self.exposed_measure = {}
        self.measure_for_current_period = {}

    def __iter__(self):
        raise NotImplementedError()

    def _init_metrics(self):
        self.energy_metric = Gauge(self.metric_name, self.metric_description, ['sensor', 'target'] + self.tags)

    def _expose_data(self, _, measure):
        kwargs = {label: measure['tags'][label] for label in measure['tags']}
        try:
            self.energy_metric.labels(**kwargs).set(measure['value'])
        except TypeError:
            self.energy_metric.labels(kwargs).set(measure['value'])

    def _report_to_measure_and_key(self, report):
        value = self.report_type.to_prometheus(report, self.tags)
        key = ''.join([str(value['tags'][tag]) for tag in value['tags']])
        return key, value

    def _update_exposed_measure(self):
        for key in self.exposed_measure:
            if key not in self.measure_for_current_period:
                args = self.exposed_measure[key]
                self.energy_metric.remove(*args)
        self.exposed_measure = self.measure_for_current_period
        self.measure_for_current_period = {}

    def save(self, report: Report):
        """
        Override from BaseDB

        :param report: Report to save
        """
        key, measure = self._report_to_measure_and_key(report)
        if self.current_ts != measure['time']:
            self.current_ts = measure['time']
            self._update_exposed_measure()

        self._expose_data(key, measure)
        if key not in self.measure_for_current_period:
            args = [measure['tags'][label] for label in measure['tags']]
            self.measure_for_current_period[key] = args

    def save_many(self, reports: List[Report]):
        """
        Save a batch of data

        :param reports: Batch of data.
        """
        for report in reports:
            self.save(report)



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Callable, List


from thespian.actors import ActorAddress

from powerapi.report import Report
from powerapi.exception import PowerAPIException


class RouterWithoutRuleException(PowerAPIException):
    """
    Exception raised when a filter route with 0 filters
    """
    def __init__(self):
        PowerAPIException.__init__(self, "")


class Filter:
    """
    Filter class

    A filter allow the Puller to route Report of the database to Dispatchers
    by fixing some rules.
    """

    def __init__(self):
        self.filters = []

    def filter(self, rule: Callable[[Report], bool], dispatcher: ActorAddress):
        """
        Define a rule for a kind of report, and send it to the dispatcher
        if the rule accept it.

        :param (func(report) -> bool) rule:      Function which return if
                                                 the report has to be send to
                                                 this dispatcher
        :param powerapi.Dispatcher dispatcher: Dispatcher we want to send the
                                                 report
        """
        self.filters.append((rule, dispatcher))

    def route(self, report: Report) -> List[ActorAddress]:
        """
        Get the list of dispatchers to whom send the report, or None

        :param powerapi.Report report: Message to send
        """
        if not self.filters:
            raise RuntimeError()

        dispatchers = []
        for rule, dispatcher in self.filters:
            if rule(report):
                dispatchers.append(dispatcher)

        return dispatchers



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import re
import os

from typing import List, Type

from powerapi.report import Report
from .base_db import BaseDB, DBError


class DirectoryDoesNotExistForVirtioFS(DBError):
    """
    Error raised when root or VM directory does not eist
    """
    def __init__(self, directory):
        DBError.__init__(self, 'No such file or directory : ' + directory)


class VirtioFSDB(BaseDB):
    """
    write power consumption of virtual machine in a file shared with the virtual machine

    The File will be written in a directory created by the VM manager.

    A regular expression must be given by the VM manager to extract vm name from target name. VM name is used to find directory that contains the output file.
    """
    def __init__(self, report_type: Type[Report], vm_name_regexp: str, root_directory_name: str,
                 vm_directory_name_prefix: str = '', vm_directory_name_suffix: str = ''):
        """
        :param vm_name_regexp:           regexp used to extract vm name from report. The regexp must match the name of the target
                                         in the HWPC-report and a group must
        :param root_directory_name:      directory where VM directory will be stored
        :param vm_directory_name_prefix: first part of the VM directory name
        :param vm_directory_name_suffix: last part of the VM directory name
        """
        BaseDB.__init__(self, report_type)

        self.vm_name_regexp = re.compile(vm_name_regexp)
        self.root_directory_name = root_directory_name
        self.vm_directory_name_suffix = vm_directory_name_suffix
        self.vm_directory_name_prefix = vm_directory_name_prefix

    def _generate_vm_directory_name(self, target):
        match = re.match(self.vm_name_regexp, target)
        if match is None:
            return None
        vm_name = match.groups()[0]
        return self.vm_directory_name_prefix + vm_name + self.vm_directory_name_suffix

    def connect(self):
        if not os.path.exists(self.root_directory_name):
            raise DirectoryDoesNotExistForVirtioFS(self.root_directory_name)

    def save(self, report: Report):
        directory_name = self._generate_vm_directory_name(report.target)
        if directory_name is None:
            return

        vm_filename, power = self.report_type.to_virtiofs_db(report)
        vm_filename_path = self.root_directory_name + '/' + directory_name + '/'
        if not os.path.exists(vm_filename_path):
            raise DirectoryDoesNotExistForVirtioFS(vm_filename_path)

        vm_file = open(vm_filename_path + vm_filename, 'w+')
        vm_file.write(str(power))
        vm_file.close()

    def save_many(self, reports: List[Report]):
        for report in reports:
            self.save(report)



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import json
import logging

from typing import Dict, Type
from powerapi.cli.parser import MainParser, ComponentSubParser

from powerapi.cli.parser import MissingValueException
from powerapi.cli.parser import BadTypeException, BadContextException, MissingArgumentException
from powerapi.cli.parser import UnknowArgException, AlreadyAddedArgumentException, AlreadyAddedSubparserException
from powerapi.cli.parser import store_val
from powerapi.cli.parser import _find_longest_name


def extract_file_names(arg, val, args, acc):
    """
    action used to convert string from --files parameter into a list of file name
    """
    acc[arg] = val.split(',')
    return args, acc


class ConfigParserArg:
    """
    Parser abstraction that retrieve the config.
    """
    def __init__(self, name_list: list, is_flag: bool, default_value, help: str, type: Type, mandatory: bool):
        self.names = name_list
        self.is_flag = is_flag
        self.default_value = default_value
        self.help = help
        self.type = type
        self.is_mandatory = mandatory


class ConfigParser:
    """ The config parser call the right parser on the CLI or the config file and then verify the type and add the default id needed."""
    def __init__(self):
        self.cli_parser = None
        self.args = {}

    def _get_mandatory_args(self):
        """
        Return the list of mandatory arguments
        """
        mand_arg = []
        for name, value in self.args.items():
            if value.is_mandatory:
                mand_arg.append(name)
        return mand_arg

    def add_argument(self, *names, flag=False, action=store_val, default=None,
                     help='', type=str, mandatory=False):
        """
        Add an argument to the parser and its specification
        """
        parser_val = ConfigParserArg(list(names), flag, default, help, type, mandatory)

        name = _find_longest_name(names)
        if name in self.args:
            raise AlreadyAddedArgumentException(name)
        self.args[name] = parser_val
        self.cli_parser.add_argument(*names, flag=flag, action=action, default=default, help=help, type=type)


class SubConfigParser(ConfigParser):
    """
    Sub Parser for MainConfigParser
    """
    def __init__(self, name: str):
        ConfigParser.__init__(self)
        self.subparser = {}
        self.cli_parser = ComponentSubParser(name)
        self.name = name

    def validate(self, conf: Dict):
        """ Check the parsed configuration"""

        # Check that all the mandatory arguments are precised
        mandatory_args = self._get_mandatory_args()
        for arg in mandatory_args:
            if arg not in conf:
                raise MissingArgumentException(arg)

        # check types
        for args, value in conf.items():
            for _, waited_value in self.args.items():
                if args in waited_value.names:
                    # check type
                    if not isinstance(value, waited_value.type) and not waited_value.is_flag:
                        raise BadTypeException(args, waited_value.type)

        for args, value in self.args.items():
            is_precised = False
            for name in value.names:
                if name in conf:
                    is_precised = True
                    break
            if not is_precised and value.default_value is not None:
                conf[args] = value.default_value

        return conf


class MainConfigParser(ConfigParser):
    """
    Parser abstraction for the configuration
    """
    def __init__(self):
        ConfigParser.__init__(self)
        self.subparser = {}
        self.cli_parser = MainParser()

    def add_subparser(self, name, subparser: SubConfigParser, help=''):

        """
        Add a SubParser to call when <name> is encoutered
        When name is encoutered, the subarpser such as subparser.name match conf[name].type

        """
        if name in self.subparser:
            if subparser.name in list(self.subparser[name]):
                raise AlreadyAddedSubparserException(name)
        else:
            self.subparser[name] = {}

        self.subparser[name][subparser.name] = subparser

        self.cli_parser.add_actor_subparser(name, subparser.cli_parser, help)

    def _parse_cli(self, cli_line):
        return self.cli_parser.parse(cli_line)

    @staticmethod
    def _parse_file(filename):
        config_file = open(filename, 'r')
        conf = json.load(config_file)
        return conf

    def _validate(self, conf: Dict):
        """ Check the parsed configuration"""

        # Check that all the mandatory arguments are precised
        mandatory_args = self._get_mandatory_args()
        for arg in mandatory_args:
            if arg not in conf:
                raise MissingArgumentException(arg)

        # check types
        for args, value in conf.items():
            is_an_arg = False
            if args in self.subparser:
                for _, dic_value in value.items():
                    self.subparser[args][dic_value["type"]].validate(dic_value)
                    is_an_arg = True

            for _, waited_value in self.args.items():
                if args in waited_value.names:
                    is_an_arg = True
                    # check type
                    if not isinstance(value, waited_value.type) and not waited_value.is_flag:
                        raise BadTypeException(args, waited_value.type)

            if not is_an_arg:
                raise UnknowArgException(args)

        for args, value in self.args.items():
            is_precised = False
            for name in value.names:
                if name in conf:
                    is_precised = True
                    break
            if not is_precised and value.default_value is not None:
                conf[args] = value.default_value

        return conf

    def parse(self, args=None):
        """
        Find the configuration method (CLI or config file)
        Call the method to produce a configuration dictionnary
        check the configuration
        """

        if args is None:
            args = sys.argv
        i = 0
        filename = None
        for s in args:
            if s == '--config-file':
                if i + 1 == len(args):
                    logging.error("CLI Error: config file path needed with argument --config-file")
                    sys.exit(-1)
                filename = args[i + 1]
            i += 1

        try:
            if filename is not None:
                conf = self._parse_file(filename)
            else:
                conf = self._parse_cli(args[1:])
            conf = self._validate(conf)

        except MissingValueException as exn:
            msg = 'CLI error: argument ' + exn.argument_name + ': expect a value'
            logging.error(msg)
            sys.exit(-1)

        except BadTypeException as exn:
            msg = "Configuration error: " + exn.msg
            logging.error(msg)
            sys.exit(-1)

        except UnknowArgException as exn:
            msg = 'Configuration error: unknow argument ' + exn.argument_name
            logging.error(msg)
            sys.exit(-1)

        except BadContextException as exn:
            msg = 'CLI error: argument ' + exn.argument_name
            msg += ' not used in the correct context\nUse it with the following arguments:'
            for main_arg_name, context_name in exn.context_list:
                msg += '\n  --' + main_arg_name + ' ' + context_name
            logging.error(msg)
            sys.exit(-1)

        except FileNotFoundError:
            logging.error("Configuration Error: configuration file not found")
            sys.exit(-1)

        except json.JSONDecodeError as exn:
            logging.error('Configuration Error: JSON Error: ' + exn.msg + ' at line' + exn.lineno + ' colomn ' + exn.colno)
            sys.exit(-1)

        except MissingArgumentException as exn:
            logging.error("Configuration Error: " + exn.msg)
            sys.exit(-1)

        return conf



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging
import sys

from typing import Type

from thespian.actors import ActorSystem

from powerapi.message import OKMessage, ErrorMessage, StartMessage, EndMessage
from powerapi.actor import InitializationException, Actor
from powerapi.exception import PowerAPIExceptionWithMessage
from powerapi.pusher import PusherActor
from powerapi.puller import PullerActor
from powerapi.dispatcher import DispatcherActor


class ActorCrashedException(PowerAPIExceptionWithMessage):
    """
    Exception raised when an actor crashed and can't be restarted
    """


class ActorLogFilter(logging.Filter):
    """
    Log filter
    """
    def filter(self, record):
        """
        filter logs that was produced by actor
        """
        return 'actor_name' in record.__dict__


class NotActorLogFilter(logging.Filter):
    """
    Log filter
    """
    def filter(self, record):
        """
        filter logs that was not produced by actor
        """
        return 'actorAddress' not in record.__dict__


LOG_DEF = {
    'version': 1,
    'formatters': {
        'normal': {'format': '%(levelname)s::%(created)s::ROOT::%(message)s'},
        'actor': {'format': '%(levelname)s::%(created)s::%(actor_name)s::%(message)s'}},
    'filters': {
        'isActorLog': {'()': ActorLogFilter},
        'notActorLog': {'()': NotActorLogFilter}},
    'handlers': {
        'h1': {'class': 'logging.StreamHandler',
               'stream': sys.stdout,
               'formatter': 'normal',
               'filters': ['notActorLog'],
               'level': logging.DEBUG},
        'h2': {
            'class': 'logging.StreamHandler',
            'stream': sys.stdout,
            'formatter': 'actor',
            'filters': ['isActorLog'],
            'level': logging.DEBUG}
    },
    'loggers': {'': {'handlers': ['h1', 'h2'], 'level': logging.DEBUG}}
}


class Supervisor:
    """
    Interface with Actor system
    Used to launch, stop and monitor actors
    """
    def __init__(self, verbose_mode: bool):
        """
        :param verbose_mode: Specify the log level : True -> DEBUG False -> INFO
        """
        if verbose_mode:
            LOG_DEF['handlers']['h1']['level'] = logging.DEBUG
            LOG_DEF['handlers']['h2']['level'] = logging.DEBUG
        else:
            LOG_DEF['handlers']['h1']['level'] = logging.INFO
            LOG_DEF['handlers']['h2']['level'] = logging.INFO

        self.system = ActorSystem(systemBase='multiprocQueueBase', logDefs=LOG_DEF)
        self.pushers = {}
        self.pullers = {}
        self.dispatchers = {}
        self.actors = {}

    def launch(self, actor_cls: Type[Actor], start_message: StartMessage):
        """
        create an actor from a given class and send it a start message.
        :param actor_cls: class used to create the actor
        :param start_message: message used to initialize the actor
        :raise InitializationException: if an error occurs during actor initialiszation process
        """
        name = start_message.name
        address = self.system.createActor(actor_cls)
        answer = self.system.ask(address, start_message)

        if isinstance(answer, OKMessage):
            self._add_actor(address, name, actor_cls)
            logging.info('launch %s actor', name)
            return address
        elif isinstance(answer, ErrorMessage):
            raise InitializationException(answer.error_message)
        raise InitializationException("Unknow message type : " + str(type(answer)))

    def _add_actor(self, address, name, actor_cls):
        if issubclass(actor_cls, PusherActor):
            self.pushers[name] = address
        elif issubclass(actor_cls, PullerActor):
            self.pullers[name] = address
        elif issubclass(actor_cls, DispatcherActor):
            self.dispatchers[name] = address
        else:
            raise AttributeError('Actor is not a DispatcherActor, PusherActor or PullerActor')
        self.actors[name] = address

    def _wait_actors(self):
        for _ in self.pushers:
            self.system.listen()

    def shutdown(self):
        """
        Shutdown the entire actor system and all the actors
        """
        self.system.shutdown()

    def monitor(self):
        """
        wait for an actor to send an EndMessage or for an actor to crash
        """
        while True:
            msg = self.system.listen(1)
            if msg is None:
                pass
            elif isinstance(msg, EndMessage):
                self._wait_actors()
                return
            else:
                logging.error("Unknow message type : %s", str(type(msg)))



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

from datetime import datetime
from typing import Dict, Any

from powerapi.report.report import Report, BadInputData, CSV_HEADER_COMMON, CsvLines


CSV_HEADER_HWPC = CSV_HEADER_COMMON + ['socket', 'cpu']


class HWPCReport(Report):
    """
    HWPCReport class

    JSON HWPC format

    .. code-block:: json

        {
         'timestamp': $int,
         'sensor': '$str',
         'target': '$str',
         'groups' : {
            '$group_name': {
               '$socket_id': {
                   '$core_id': {
                       '$event_name': '$int',
                       ...
                   }
                   ...
               }
               ...
            }
            ...
         }
        }
    """

    def __init__(self, timestamp: datetime, sensor: str, target: str, groups: Dict[str, Dict], metadata: Dict[str, Any] = {}):
        """
        Initialize an HWPC report using the given parameters.
        :param datetime timestamp: Timestamp of the report
        :param str sensor: Sensor name
        :param str target: Target name
        :param Dict groups: Events groups
        """
        Report.__init__(self, timestamp, sensor, target, metadata)

        #: (dict): Events groups
        self.groups = groups

    def __repr__(self) -> str:
        return 'HWCPReport(%s, %s, %s, %s, %s)' % (self.timestamp, self.sensor, self.target, sorted(self.groups.keys()), str(self.metadata))

    @staticmethod
    def from_json(data: Dict) -> HWPCReport:
        """
        Generate a report using the given data.
        :param data: Dictionary containing the report attributes
        :return: The HWPC report initialized with the given data
        """
        try:
            ts = Report._extract_timestamp(data['timestamp'])
            metadata = {} if 'metadata' not in data else data['metadata']
            return HWPCReport(ts, data['sensor'], data['target'], data['groups'], metadata)
        except KeyError as exn:
            raise BadInputData('HWPC report require field ' + str(exn.args[0]) + ' in json document', data) from exn
        except ValueError as exn:
            raise BadInputData(exn.args[0], data) from exn

    @staticmethod
    def to_json(report: HWPCReport) -> Dict:
        return report.__dict__

    @staticmethod
    def from_mongodb(data: Dict) -> HWPCReport:
        """
        :return: a HWPCReport from a dictionary pulled from mongodb
        """
        return HWPCReport.from_json(data)

    @staticmethod
    def to_mongodb(report: HWPCReport) -> Dict:
        """
        :return: a dictionary, that can be stored into a mongodb, from a given HWPCReport
        """
        return HWPCReport.to_json(report)

    @staticmethod
    def from_csv_lines(lines: CsvLines) -> HWPCReport:
        """
        :param lines: list of pre-parsed lines. a line is a tuple composed with :
                         - the file name where the line were read
                         - a dictionary where key is column name and value is the value read from the line
        :return: a HWPCReport that contains value from the given lines
        """
        sensor_name = None
        target = None
        timestamp = None
        groups = {}
        metadata = {}

        for file_name, row in lines:
            group_name = file_name[:-4] if file_name[len(file_name) - 4:] == '.csv' else file_name
            try:
                if sensor_name is None:
                    sensor_name = row['sensor']
                    target = row['target']
                    timestamp = HWPCReport._extract_timestamp(row['timestamp'])
                else:
                    if sensor_name != row['sensor']:
                        raise BadInputData('csv line with different sensor name are mixed into one report', row)
                    if target != row['target']:
                        raise BadInputData('csv line with different target are mixed into one report', row)
                    if timestamp != HWPCReport._extract_timestamp(row['timestamp']):
                        raise BadInputData('csv line with different timestamp are mixed into one report', row)

                HWPCReport._create_group(row, groups, group_name)

                for key, value in row.items():
                    if key not in CSV_HEADER_HWPC:
                        metadata[key] = value

            except KeyError as exn:
                raise BadInputData('missing field ' + str(exn.args[0]) + ' in csv file ' + file_name, row) from exn
            except ValueError as exn:
                raise BadInputData(exn.args[0], row) from exn

        return HWPCReport(timestamp, sensor_name, target, groups, metadata)

    @staticmethod
    def _create_group(row, groups, group_name):

        if group_name not in groups:
            groups[group_name] = {}

        if row['socket'] not in groups[group_name]:
            groups[group_name][row['socket']] = {}

        if row['cpu'] not in groups[group_name][row['socket']]:
            groups[group_name][row['socket']][row['cpu']] = {}



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations
from typing import Type, List
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from powerapi.database import BaseDB
    from powerapi.filter import Filter
    from powerapi.dispatcher import RouteTable
    from powerapi.formula import FormulaActor, FormulaValues, DomainValues
    from powerapi.report_modifier import ReportModifier


class Message:
    """
    Abstract Message class
    """
    def __init__(self, sender_name: str):
        self.sender_name = sender_name

    def __str__(self):
        raise NotImplementedError()


class PingMessage(Message):
    """
    Message used to test if an actor is alive
    """
    def __init__(self, sender_name: str):
        Message.__init__(self, sender_name)

    def __str__(self):
        return "PingMessage"


class OKMessage(Message):
    """
    Message send to acknowledge last received message
    """
    def __init__(self, sender_name: str):
        Message.__init__(self, sender_name)

    def __str__(self):
        return "OKMessage"


class ErrorMessage(Message):
    """
    Message used to indicate that an error as occuried
    """

    def __init__(self, sender_name: str, error_message: str):
        """
        :param str error_code: message associated to the error
        """
        Message.__init__(self, sender_name)
        self.error_message = error_message

    def __str__(self):
        return "ErrorMessage : " + self.error_message


class StartMessage(Message):
    """
    Message that ask the actor to launch its initialisation process
    """
    def __init__(self, sender_name: str, name: str):
        Message.__init__(self, sender_name)
        self.name = name

    def __str__(self):
        return "StartMessage"


class EndMessage(Message):
    """
    Message sent by actor to its parent when it terminate itself
    """
    def __init__(self, sender_name: str):
        Message.__init__(self, sender_name)

    def __str__(self):
        return "EndMessage"


class PullerStartMessage(StartMessage):
    """
    Message used to start a Puller actor
    """
    def __init__(self, sender_name: str, name: str, database: BaseDB, report_filter: Filter, stream_mode: bool, report_modifiers: List[ReportModifier] = []):
        """
        :param sender_name: name of the actor that send the message
        :param name: puller actor name
        :param database: database used by the puller to pull report
        :param report_filter: report filter used to filter reports
        :param stream_mode: True if stream mode is enabled
        :param report_modifier_list: list of ReportModifier used to modify report before sending them to dispatcher
        """
        StartMessage.__init__(self, sender_name, name)
        self.database = database
        self.report_filter = report_filter
        self.stream_mode = stream_mode
        self.report_modifier_list = report_modifiers


class DispatcherStartMessage(StartMessage):
    """
    Message used to start Dispatcher actor
    """
    def __init__(self, sender_name: str, name: str, formula_class: Type[FormulaActor], formula_values: FormulaValues, route_table: RouteTable, device_id: str):
        """
        :param sender_name: name of the actor that send the message
        :param name: puller actor name
        :param formula_class: Class of the formula the dispatcher handle
        :param formula_values: Values that will be always passed to formula for initialization
        :param route_table: Dispatcher's Route table
        :param device_id: name of the device the dispatcher handle
        """
        StartMessage.__init__(self, sender_name, name)
        self.formula_class = formula_class
        self.formula_values = formula_values
        self.route_table = route_table
        self.device_id = device_id


class FormulaStartMessage(StartMessage):
    """
    Message used to start formula actor
    """
    def __init__(self, sender_name: str, name: str, formula_values: FormulaValues, domain_values: DomainValues):
        """
        :param sender_name: name of the actor that send the message
        :param name: puller actor name
        :param formula_values: values used by formula for initialization (depend of the formula class)
        :param domain_values: domain of the formula (depend of the formula class)
        """
        StartMessage.__init__(self, sender_name, name)
        self.values = formula_values
        self.domain_values = domain_values


class PusherStartMessage(StartMessage):
    """
    Message used to start a Pusher actor
    """
    def __init__(self, sender_name: str, name: str, database: BaseDB):
        """
        :param sender_name: name of the actor that send the message
        :param name: puller actor name
        :param BaseDB database: Database use for saving data.
        """
        StartMessage.__init__(self, sender_name, name)
        self.database = database



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging

from typing import Any, Type
from datetime import timedelta

from thespian.actors import ActorTypeDispatcher, ActorAddress, ActorExitRequest, WakeupMessage

from powerapi.message import PingMessage, OKMessage, ErrorMessage, StartMessage
from powerapi.exception import PowerAPIExceptionWithMessage, PowerAPIException


class InitializationException(PowerAPIExceptionWithMessage):
    """
    Exception raised when an actor failed to initialize itself
    """


class ActorNotInitializedException(PowerAPIException):
    """
    Exception raised when a non initialized actor received a message that is not a StartMessage
    """


class Actor(ActorTypeDispatcher):
    """
    PowerAPI Actor Abstract Class
    """
    def __init__(self, start_message_cls: Type[StartMessage]):
        """
        This method must be used only in a daughter class constructor

        :param start_message_cls: Type of the message that will be used to initialize the actor
        """
        ActorTypeDispatcher.__init__(self)
        self.name: str = 'no_name_yet'
        self.parent: ActorAddress = None
        self.initialized: bool = False
        self.start_message_cls = start_message_cls

    def log_critical(self, message: str):
        """
        send a critical message to the log actor
        """
        extra_args = {'actor_name': self.name}
        logging.critical(message, extra=extra_args)

    def log_error(self, message: str):
        """
        send an error message to the log actor
        """
        extra_args = {'actor_name': self.name}
        logging.error(message, extra=extra_args)

    def log_warning(self, message: str):
        """
        send a warning message to the log actor
        """
        extra_args = {'actor_name': self.name}
        logging.warning(message, extra=extra_args)

    def log_info(self, message: str):
        """
        send an information message to the log actor
        """
        extra_args = {'actor_name': self.name}
        logging.info(message, extra=extra_args)

    def log_debug(self, message: str):
        """
        send a debug message to the log actor
        """
        extra_args = {'actor_name': self.name}
        logging.debug(message, extra=extra_args)

    def receiveMsg_PingMessage(self, message: PingMessage, sender: ActorAddress):
        """
        When receiving a PingMessage, the actor answer with a OKMessage to its sender
        """
        self.log_debug('received message ' + str(message))
        self.send(sender, OKMessage(self.name))

    def receiveMsg_StartMessage(self, message: StartMessage, sender: ActorAddress):
        """
        When receiving a StartMessage :
          - if the actor is already initialized, answer with a ErrorMessage
          - otherwise initialize the actor with the abstract method _initialization and answer with an OKMessage
        """
        self.name = message.name
        self.log_debug('received message ' + str(message))
        self.parent = sender

        if self.initialized:
            self.send(sender, ErrorMessage(self.name, 'Actor already initialized'))
            return

        if not isinstance(message, self.start_message_cls):
            self.send(sender, ErrorMessage(self.name, 'use ' + self.start_message_cls.__name__ + ' instead of StartMessage'))
            return

        try:
            self._initialization(message)
        except InitializationException as exn:
            self.send(sender, ErrorMessage(self.name, exn.msg))
            self.send(self.myAddress, ActorExitRequest())
            return

        self.initialized = True
        self.send(sender, OKMessage(self.name))
        self.log_info(self.name + ' started')

    def receiveMsg_ErrorMessage(self, message: ErrorMessage, _: ActorAddress):
        """
        When receiving an error message send it to the log actor
        """
        self.log_error(str(message.error_message))

    def receiveUnrecognizedMessage(self, message: Any, sender: ActorAddress):
        """
        When receiving a message with a type that can't be handle, the actor answer with an ErrorMessage
        """
        self.log_warning('received unrecognized message : ' + str(message))
        self.send(sender, ErrorMessage(self.name, "did not recognize the message type : " + str(type(message))))

    def receiveMsg_ActorExitRequest(self, message: ActorExitRequest, _: ActorAddress):
        """
        When receive ActorExitRequestMessage log it and exit
        """
        self.log_debug('received message ' + str(message))
        self.log_info(self.name + ' exit')

    def _initialization(self, start_message: StartMessage):
        pass


class TimedActor(Actor):
    """
    An actor that process a task at regular time interval
    """

    def __init__(self, start_message_cls, time_interval: float):
        """
        :param time_interval: time (in seconds) to wait between each task to launch
        """
        Actor.__init__(self, start_message_cls)
        self._time_interval = timedelta(seconds=time_interval)

    def receiveMsg_StartMessage(self, message: StartMessage, sender: ActorAddress):
        """
        When receiving a StartMessage, initialize the actor and set it into sleeping phase
        ask the actor system to wake agter a given time period
        """
        Actor.receiveMsg_StartMessage(self, message, sender)
        self.wakeupAfter(self._time_interval)

    def receiveMsg_WakeupMessage(self, _: WakeupMessage, __: ActorAddress):
        """
        When receiving a WakeupMessage, launch the actor task
        """
        if self.initialized:
            self._launch_task()
        else:
            self.log_error('received a wakeup message without being initialized before')
            raise ActorNotInitializedException()

    def _launch_task(self):
        raise NotImplementedError()



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging
import os

from typing import Dict


class ConfigValidator:
    """
    Validate powerapi config and initialize missing default values
    """
    @staticmethod
    def validate(config: Dict):
        """
        Validate powerapi config and initialize missing default values
        """
        if 'verbose' not in config:
            config['verbose'] = logging.NOTSET
        if 'stream' not in config:
            config['stream'] = False
        if 'output' not in config:
            logging.error("no output configuration found")
            return False

        for output_type in config['output']:
            output_config = config['output'][output_type]
            if 'model' not in output_config:
                output_config['model'] = 'HWPCReport'
            if 'name' not in output_config:
                output_config['name'] = 'default_pusher'

        if 'input' not in config:
            logging.error("no input configuration found")
            return False

        for input_type in config['input']:
            input_config = config['input'][input_type]
            if 'model' not in input_config:
                input_config['model'] = 'PowerReport'
            if 'name' not in input_config:
                input_config['name'] = 'default_puller'

        if not ConfigValidator._validate_input(config):
            return False
        return True

    @staticmethod
    def _validate_input(config: Dict):
        for _, input_config in config['input'].items():
            if input_config['type'] == 'csv':
                for file_name in input_config['files']:
                    if not os.access(file_name, os.R_OK):
                        return False
        return True



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from datetime import datetime
import logging
from typing import List, Type
import os
import json
from powerapi.database.base_db import BaseDB, DBError, IterDB
from powerapi.report import Report


class FileBadDBError(DBError):
    """
    Error raised when hostname/port fail
    """

    def __init__(self, filename):
        DBError.__init__(self, "File error : " + filename + " not found")


class FileIterDB(IterDB):
    """
    FileIterDB class

    Class for iterating in a file
    """

    def __init__(self, db, report_type, stream_mode, filename):
        """ """
        IterDB.__init__(self, db, report_type, stream_mode)
        self.previousJson = ""
        self.__iter__()
        self.filename = filename

    def __iter__(self):
        """
        Create the iterator for get the data
        """
        return self

    def __next__(self) -> Report:
        """
        Allow to get the next data
        :raise: StopIteration in stream mode when no report was found.
        """

        file_object = open(self.filename, "r")
        json_str = file_object.read()
        file_object.close()

        if json_str is None:
            raise StopIteration()

        if json_str == self.previousJson:
            logging.error("Error : Report did not change since last read")
            raise StopIteration()

        self.previousJson = json_str

        return self.report_type.from_json(json.loads(json_str))


class FileDB(BaseDB):
    """
    FileDB class herited from BaseDB

    Allow to handle a FileDB database in reading or writing.
    """

    def __init__(self, report_type: Type[Report], filename: str):
        """
        :param report_type:        Type of the report handled by this database
        :param filename:        Name of the file containing the report
        """

        BaseDB.__init__(self, report_type)

        self.filename = filename

    def connect(self):
        """
        Override from BaseDB.

        It check that the file exist
        """
        if not os.path.exists(self.filename):
            raise FileBadDBError(self.filename)

    def iter(self, stream_mode: bool) -> FileIterDB:
        """
        Create the iterator for get the data
        """
        return FileIterDB(self, self.report_type, stream_mode, self.filename)

    def save(self, report: Report):
        """
        Override from BaseDB

        :param report: Report to save
        """
        file_object = open(self.filename, "w")
        line = {
            "sensor": report.sensor,
            "target": report.target,
            "timestamp": int(datetime.timestamp(report.timestamp) * 1000),
            "power": report.power,
        }

        final_dict = {"PowerReport": [line]}

        file_object.truncate(0)
        file_object.write(str(final_dict))
        file_object.close()

    def save_many(self, reports: List[Report]):
        """
        Allow to save a batch of data

        :param reports: Batch of data.
        """
        for report in reports:
            self.save(report)



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import pytest
import pymongo


MONGO_URI = "mongodb://localhost:27017/"
MONGO_INPUT_COLLECTION_NAME = 'test_input'
MONGO_OUTPUT_COLLECTION_NAME = 'test_output'
MONGO_DATABASE_NAME = 'MongoDB1'


@pytest.fixture
def mongo_database(mongodb_content):
    """
    connect to a local mongo database (localhost:27017) and store data contained in the list influxdb_content
    after test end, delete the data
    """
    _gen_base_db_test(MONGO_URI, mongodb_content)
    yield None
    _clean_base_db_test(MONGO_URI)


def _gen_base_db_test(uri, content):
    mongo = pymongo.MongoClient(uri)
    db = mongo[MONGO_DATABASE_NAME]

    # delete collection if it already exist
    db[MONGO_INPUT_COLLECTION_NAME].drop()
    db.create_collection(MONGO_INPUT_COLLECTION_NAME)
    for item in content:
        db[MONGO_INPUT_COLLECTION_NAME].insert_one(item)

    # delete output collection
    db[MONGO_OUTPUT_COLLECTION_NAME].drop()
    mongo.close()


def _clean_base_db_test(uri):
    """
    drop test_hwrep and test_result collections
    """
    mongo = pymongo.MongoClient(uri)
    db = mongo[MONGO_DATABASE_NAME]
    db[MONGO_INPUT_COLLECTION_NAME].drop()
    db[MONGO_OUTPUT_COLLECTION_NAME].drop()
    mongo.close()
